@@ -1,3 +1,17 @@
+/*
+ * Copyright (C) 1996-2005 The Free Software Foundation, Inc.
+ *
+ * This program is free software; you can redistribute it and/or modify
+ * it under the terms of the GNU General Public License as published by
+ * the Free Software Foundation; either version 2, or (at your option)
+ * any later version.
+ *
+ * This program is distributed in the hope that it will be useful,
+ * but WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ * GNU General Public License for more details.
+ */
+
 /* Code for the buffer data structure.  */
 
 #include <assert.h>
@@ -60,6 +74,11 @@ void
 buf_free (buf)
      struct buffer *buf;
 {
+    if (buf->closure != NULL)
+    {
+	free (buf->closure);
+	buf->closure = NULL;
+    }
     if (buf->data != NULL)
     {
 	buf->last->next = free_buffer_data;
@@ -105,11 +124,13 @@ allocate_buffer_datas ()
     /* Allocate buffer_data structures in blocks of 16.  */
 #define ALLOC_COUNT (16)
 
-    alc = ((struct buffer_data *)
-	   malloc (ALLOC_COUNT * sizeof (struct buffer_data)));
+    alc = xmalloc (ALLOC_COUNT * sizeof (struct buffer_data));
     space = (char *) valloc (ALLOC_COUNT * BUFFER_DATA_SIZE);
-    if (alc == NULL || space == NULL)
+    if (!space)
+    {
+	free (alc);
 	return;
+    }
     for (i = 0; i < ALLOC_COUNT; i++, alc++, space += BUFFER_DATA_SIZE)
     {
 	alc->next = free_buffer_data;
@@ -137,8 +158,23 @@ get_buffer_data ()
     return ret;
 }
 
-/* See whether a buffer is empty.  */
 
+
+/* See whether a buffer and its file descriptor is empty.  */
+int
+buf_empty (buf)
+    struct buffer *buf;
+{
+	/* Try and read any data on the file descriptor first.
+	 * We already know the descriptor is non-blocking.
+	 */
+	buf_input_data (buf, NULL);
+	return buf_empty_p (buf);
+}
+
+
+
+/* See whether a buffer is empty.  */
 int
 buf_empty_p (buf)
     struct buffer *buf;
@@ -151,10 +187,12 @@ buf_empty_p (buf)
     return 1;
 }
 
+
+
 #ifdef SERVER_FLOWCONTROL
 /*
  * Count how much data is stored in the buffer..
- * Note that each buffer is a malloc'ed chunk BUFFER_DATA_SIZE.
+ * Note that each buffer is a xmalloc'ed chunk BUFFER_DATA_SIZE.
  */
 
 int
@@ -794,7 +832,7 @@ buf_read_line (buf, line, lenp)
 	    char *p;
 	    struct buffer_data *nldata;
 
-	    p = malloc (len + 1);
+	    p = xmalloc (len + 1);
 	    if (p == NULL)
 		return -2;
 	    *line = p;
@@ -1208,6 +1246,8 @@ buf_shutdown (buf)
     return 0;
 }
 
+
+
 /* The simplest type of buffer is one built on top of a stdio FILE.
    For simplicity, and because it is all that is required, we do not
    implement setting this type of buffer into nonblocking mode.  The
@@ -1218,14 +1258,17 @@ static int stdio_buffer_output PROTO((void *, const char *, int, int *));
 static int stdio_buffer_flush PROTO((void *));
 static int stdio_buffer_shutdown PROTO((struct buffer *buf));
 
-/* Initialize a buffer built on a stdio FILE.  */
 
+
+/* Initialize a buffer built on a stdio FILE.  */
 struct stdio_buffer_closure
 {
     FILE *fp;
     int child_pid;
 };
 
+
+
 struct buffer *
 stdio_buffer_initialize (fp, child_pid, input, memory)
      FILE *fp;
@@ -1233,7 +1276,7 @@ stdio_buffer_initialize (fp, child_pid, input, memory)
      int input;
      void (*memory) PROTO((struct buffer *));
 {
-    struct stdio_buffer_closure *bc = malloc (sizeof (*bc));
+    struct stdio_buffer_closure *bc = xmalloc (sizeof (*bc));
 
     bc->fp = fp;
     bc->child_pid = child_pid;
@@ -1247,6 +1290,20 @@ stdio_buffer_initialize (fp, child_pid, input, memory)
 			   (void *) bc);
 }
 
+/* Return the file associated with a stdio buffer. */
+FILE *
+stdio_buffer_get_file (buf)
+    struct buffer *buf;
+{
+    struct stdio_buffer_closure *bc;
+
+    assert(buf->shutdown == stdio_buffer_shutdown);
+
+    bc = (struct stdio_buffer_closure *) buf->closure;
+
+    return(bc->fp);
+}
+
 /* The buffer input function for a buffer built on a stdio FILE.  */
 
 static int
@@ -1337,8 +1394,9 @@ stdio_buffer_output (closure, data, have, wrote)
     return 0;
 }
 
-/* The buffer flush function for a buffer built on a stdio FILE.  */
 
+
+/* The buffer flush function for a buffer built on a stdio FILE.  */
 static int
 stdio_buffer_flush (closure)
      void *closure;
@@ -1362,12 +1420,12 @@ static int
 stdio_buffer_shutdown (buf)
     struct buffer *buf;
 {
-    struct stdio_buffer_closure *bc = (struct stdio_buffer_closure *) buf->closure;
+    struct stdio_buffer_closure *bc = buf->closure;
     struct stat s;
     int closefp = 1;
 
     /* Must be a pipe or a socket.  What could go wrong? */
-    assert (fstat ( fileno (bc->fp), &s ) != -1);
+    assert (fstat (fileno (bc->fp), &s) != -1);
 
     /* Flush the buffer if we can */
     if (buf->flush)
@@ -1378,31 +1436,11 @@ stdio_buffer_shutdown (buf)
 
     if (buf->input)
     {
-	if (! buf_empty_p (buf)
-	    || getc (bc->fp) != EOF)
-	{
-# ifdef SERVER_SUPPORT
-	    if (server_active)
-		/* FIXME: This should probably be sysloged since it doesn't
-		 * have anywhere else to go at this point.
-		 */
-		error (0, 0, "dying gasps from client unexpected");
-	    else
-#endif
-		error (0, 0, "dying gasps from %s unexpected", current_parsed_root->hostname);
-	}
-	else if (ferror (bc->fp))
-	{
-# ifdef SERVER_SUPPORT
-	    if (server_active)
-		/* FIXME: This should probably be sysloged since it doesn't
-		 * have anywhere else to go at this point.
-		 */
-		error (0, errno, "reading from client");
-	    else
-#endif
-		error (0, errno, "reading from %s", current_parsed_root->hostname);
-	}
+	/* There used to be a check here for unread data in the buffer of on
+	 * the pipe, but it was deemed unnecessary and possibly dangerous.  In
+	 * some sense it could be second-guessing the caller who requested it
+	 * closed, as well.
+	 */
 
 # ifdef SHUTDOWN_SERVER
 	if (current_parsed_root->method != server_method)
@@ -1410,8 +1448,8 @@ stdio_buffer_shutdown (buf)
 # ifndef NO_SOCKET_TO_FD
 	{
 	    /* shutdown() sockets */
-	    if (S_ISSOCK(s.st_mode))
-		shutdown ( fileno (bc->fp), 0);
+	    if (S_ISSOCK (s.st_mode))
+		shutdown (fileno (bc->fp), 0);
 	}
 # endif /* NO_SOCKET_TO_FD */
 # ifdef START_RSH_WITH_POPEN_RW
@@ -1433,13 +1471,13 @@ stdio_buffer_shutdown (buf)
 	 * SHUTDOWN_SERVER_OUTPUT
 	 */
 	if (current_parsed_root->method == server_method)
-	    SHUTDOWN_SERVER ( fileno (bc->fp) );
+	    SHUTDOWN_SERVER (fileno (bc->fp));
 	else
 # endif
 # ifndef NO_SOCKET_TO_FD
 	/* shutdown() sockets */
-	if (S_ISSOCK(s.st_mode))
-	    shutdown ( fileno (bc->fp), 1);
+	if (S_ISSOCK (s.st_mode))
+	    shutdown (fileno (bc->fp), 1);
 # else
 	{
 	/* I'm not sure I like this empty block, but the alternative
@@ -1452,15 +1490,30 @@ stdio_buffer_shutdown (buf)
     }
 
     if (closefp && fclose (bc->fp) == EOF)
-	error (1, errno,
-	       "closing down connection to %s",
-	       current_parsed_root->hostname);
+    {
+	if (server_active)
+	{
+            /* Syslog this? */
+	}
+# ifdef CLIENT_SUPPORT
+	else
+            error (1, errno,
+                   "closing down connection to %s",
+                   current_parsed_root->hostname);
+# endif /* CLIENT_SUPPORT */
+    }
 
     /* If we were talking to a process, make sure it exited */
-    if (bc->child_pid
-	&& waitpid (bc->child_pid, (int *) 0, 0) == -1)
-	error (1, errno, "waiting for process %d", bc->child_pid);
+    if (bc->child_pid)
+    {
+	int w;
 
+	do
+	    w = waitpid (bc->child_pid, (int *) 0, 0);
+	while (w == -1 && errno == EINTR);
+	if (w == -1)
+	    error (1, errno, "waiting for process %d", bc->child_pid);
+    }
     return 0;
 }
 
@@ -1664,7 +1717,7 @@ packetizing_buffer_input (closure, data, need, size, got)
 	    /* We didn't allocate enough space in the initialize
                function.  */
 
-	    n = realloc (pb->holdbuf, count + 2);
+	    n = xrealloc (pb->holdbuf, count + 2);
 	    if (n == NULL)
 	    {
 		(*pb->buf->memory_error) (pb->buf);
@@ -1726,7 +1779,7 @@ packetizing_buffer_input (closure, data, need, size, got)
 	    outbuf = stackoutbuf;
 	else
 	{
-	    outbuf = malloc (count);
+	    outbuf = xmalloc (count);
 	    if (outbuf == NULL)
 	    {
 		(*pb->buf->memory_error) (pb->buf);
@@ -1798,7 +1851,7 @@ packetizing_buffer_output (closure, data, have, wrote)
 
     if (have > BUFFER_DATA_SIZE)
     {
-	/* It would be easy to malloc a buffer, but I don't think this
+	/* It would be easy to xmalloc a buffer, but I don't think this
            case can ever arise.  */
 	abort ();
     }
@@ -1859,8 +1912,9 @@ packetizing_buffer_output (closure, data, have, wrote)
     return buf_send_output (pb->buf);
 }
 
-/* Flush data to a packetizing buffer.  */
 
+
+/* Flush data to a packetizing buffer.  */
 static int
 packetizing_buffer_flush (closure)
      void *closure;
@@ -1874,8 +1928,9 @@ packetizing_buffer_flush (closure)
     return buf_flush (pb->buf, 0);
 }
 
-/* The block routine for a packetizing buffer.  */
 
+
+/* The block routine for a packetizing buffer.  */
 static int
 packetizing_buffer_block (closure, block)
      void *closure;
