@@ -1,9 +1,18 @@
 /*
+ * Copyright (C) 1994-2005 The Free Software Foundation, Inc.
  *
- *    You may distribute under the terms of the GNU General Public License
- *    as specified in the README file that comes with the CVS 1.0 kit.
+ * This program is free software; you can redistribute it and/or modify
+ * it under the terms of the GNU General Public License as published by
+ * the Free Software Foundation; either version 2, or (at your option)
+ * any later version.
  *
- * **************** History of Users and Module ****************
+ * This program is distributed in the hope that it will be useful,
+ * but WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ * GNU General Public License for more details.
+ */
+
+/* **************** History of Users and Module ****************
  *
  * LOGGING:  Append record to "${CVSROOT}/CVSROOTADM/CVSROOTADM_HISTORY".
  *
@@ -21,6 +30,7 @@
  *		F	"Release" cmd.
  *		W	"Update" cmd - No User file, Remove from Entries file.
  *		U	"Update" cmd - File was checked out over User file.
+ *		P	"Update" cmd - User file was patched.
  *		G	"Update" cmd - File was merged successfully.
  *		C	"Update" cmd - File was merged and shows overlaps.
  *		M	"Commit" cmd - "Modified" file.
@@ -34,9 +44,9 @@
  *
  *  CurDir	The directory where the action occurred.  This should be the
  *		absolute path of the directory which is at the same level as
- *		the "Repository" field (for W,U,G,C & M,A,R).
+ *		the "Repository" field (for W,U,P,G,C & M,A,R).
  *
- *  Repository	For record types [W,U,G,C,M,A,R] this field holds the
+ *  Repository	For record types [W,U,P,G,C,M,A,R] this field holds the
  *		repository read from the administrative data where the
  *		command was typed.
  *		T	"A" --> New Tag, "D" --> Delete Tag
@@ -48,11 +58,11 @@
  *		O,E	The Tag or Date, if specified, else "" (null field).
  *		F	"" (null field)
  *		W	The Tag or Date, if specified, else "" (null field).
- *		U	The Revision checked out over the User file.
+ *		U,P	The Revision checked out over the User file.
  *		G,C	The Revision(s) involved in merge.
  *		M,A,R	RCS Revision affected.
  *
- *  argument	The module (for [TOEUF]) or file (for [WUGCMAR]) affected.
+ *  argument	The module (for [TOEF]) or file (for [WUPGCMAR]) affected.
  *
  *
  *** Report categories: "User" and "Since" modifiers apply to all reports.
@@ -60,7 +70,7 @@
  *
  *   Extract list of record types
  *
- *	-e, -x [TOEFWUGCMAR]
+ *	-e, -x [TOEFWUPGCMAR]
  *
  *		Extracted records are simply printed, No analysis is performed.
  *		All "field" modifiers apply.  -e chooses all types.
@@ -93,7 +103,7 @@
  *		modules are remembered.  Only records matching exactly those
  *		files and repositories are shown.  Sorting by "module", then
  *		filename, is implied.  If -l ("last modified") is specified,
- *		then "update" records (types WUCG), tag and release records
+ *		then "update" records (types WUPCG), tag and release records
  *		are ignored and the last (by date) "modified" record.
  *
  *   TAG history
@@ -170,7 +180,7 @@
  *	cvs hi -e -u user
  *
  *** Dump (eXtract) specified record types
- *	cvs hi -x [TOFWUGCMAR]
+ *	cvs hi -x [TOEFWUPGCMAR]
  *
  *
  * FUTURE:		J[Join], I[Import]  (Not currently implemented.)
@@ -178,6 +188,7 @@
  */
 
 #include "cvs.h"
+#include "history.h"
 #include "savecwd.h"
 
 static struct hrec
@@ -208,7 +219,6 @@ static void save_file PROTO((char *dir, char *name, char *module));
 static void save_module PROTO((char *module));
 static void save_user PROTO((char *name));
 
-#define ALL_REC_TYPES "TOEFWUCGMAR"
 #define USER_INCREMENT	2
 #define FILE_INCREMENT	128
 #define MODULE_INCREMENT 5
@@ -217,6 +227,7 @@ static void save_user PROTO((char *name));
 static short report_count;
 
 static short extract;
+static short extract_all;
 static short v_checkout;
 static short modified;
 static short tag_report;
@@ -234,7 +245,7 @@ static short tz_local;
 static time_t tz_seconds_east_of_GMT;
 static char *tz_name = "+0000";
 
-char *logHistory = ALL_REC_TYPES;
+char *logHistory;
 
 /* -r, -t, or -b options, malloc'd.  These are "" if the option in
    question is not specified or is overridden by another option.  The
@@ -290,7 +301,7 @@ static const char *const history_usg[] =
     "        -c              Committed (Modified) files\n",
     "        -o              Checked out modules\n",
     "        -m <module>     Look for specified module (repeatable)\n",
-    "        -x [TOEFWUCGMAR] Extract by record type\n",
+    "        -x [" ALL_HISTORY_REC_TYPES "] Extract by record type\n",
     "        -e              Everything (same as -x, but all record types)\n",
     "   Flags:\n",
     "        -a              All users (Default is self)\n",
@@ -399,9 +410,9 @@ history (argc, argv)
 		break;
 	    case 'e':
 		report_count++;
-		extract++;
+		extract_all++;
 		free (rec_types);
-		rec_types = xstrdup (ALL_REC_TYPES);
+		rec_types = xstrdup (ALL_HISTORY_REC_TYPES);
 		break;
 	    case 'l':			/* Find Last file record */
 		last_entry = 1;
@@ -414,8 +425,11 @@ history (argc, argv)
 		working = 1;
 		break;
 	    case 'X':			/* Undocumented debugging flag */
+#ifdef DEBUG
 		histfile = optarg;
+#endif
 		break;
+
 	    case 'D':			/* Since specified date */
 		if (*since_rev || *since_tag || *backto)
 		{
@@ -437,7 +451,7 @@ history (argc, argv)
 		backto = xstrdup (optarg);
 		break;
 	    case 'f':			/* For specified file */
-		save_file ("", optarg, (char *) NULL);
+		save_file (NULL, optarg, NULL);
 		break;
 	    case 'm':			/* Full module report */
 		if (!module_report++) report_count++;
@@ -446,7 +460,7 @@ history (argc, argv)
 		save_module (optarg);
 		break;
 	    case 'p':			/* For specified directory */
-		save_file (optarg, "", (char *) NULL);
+		save_file (optarg, NULL, NULL);
 		break;
 	    case 'r':			/* Since specified Tag/Rev */
 		if (since_date || *since_tag || *backto)
@@ -482,7 +496,7 @@ history (argc, argv)
 		    char *cp;
 
 		    for (cp = optarg; *cp; cp++)
-			if (!strchr (ALL_REC_TYPES, *cp))
+			if (!strchr (ALL_HISTORY_REC_TYPES, *cp))
 			    error (1, 0, "%c is not a valid report type", *cp);
 		}
 		free (rec_types);
@@ -529,7 +543,7 @@ history (argc, argv)
     argc -= optind;
     argv += optind;
     for (i = 0; i < argc; i++)
-	save_file ("", argv[i], (char *) NULL);
+	save_file (NULL, argv[i], NULL);
 
 
     /* ================ Now analyze the arguments a bit */
@@ -584,6 +598,8 @@ history (argc, argv)
 	    option_with_arg ("-t", since_tag);
 	for (mod = user_list; mod < &user_list[user_count]; ++mod)
 	    option_with_arg ("-u", *mod);
+	if (extract_all)
+	    send_arg("-e");
 	if (extract)
 	    option_with_arg ("-x", rec_types);
 	option_with_arg ("-z", tz_name);
@@ -607,7 +623,7 @@ history (argc, argv)
 	    (void) strcat (rec_types, "T");
 	}
     }
-    else if (extract)
+    else if (extract || extract_all)
     {
 	if (user_list)
 	    user_sort++;
@@ -636,7 +652,7 @@ history (argc, argv)
     else if (module_report)
     {
 	free (rec_types);
-	rec_types = xstrdup (last_entry ? "OMAR" : ALL_REC_TYPES);
+	rec_types = xstrdup (last_entry ? "OMAR" : ALL_HISTORY_REC_TYPES);
 	module_sort++;
 	repos_sort++;
 	file_sort++;
@@ -697,31 +713,44 @@ history (argc, argv)
 void
 history_write (type, update_dir, revs, name, repository)
     int type;
-    char *update_dir;
-    char *revs;
-    char *name;
-    char *repository;
+    const char *update_dir;
+    const char *revs;
+    const char *name;
+    const char *repository;
 {
     char *fname;
     char *workdir;
     char *username = getcaller ();
     int fd;
     char *line;
-    char *slash = "", *cp, *cp2, *repos;
+    char *slash = "", *cp;
+    const char *cp2, *repos;
     int i;
     static char *tilde = "";
     static char *PrCurDir = NULL;
 
-    if (logoff)			/* History is turned off by cmd line switch */
+    if (logoff)			/* History is turned off by noexec or
+				 * readonlyfs.
+				 */
 	return;
-    if ( strchr(logHistory, type) == NULL )	
+    if (strchr (logHistory, type) == NULL)
 	return;
-    fname = xmalloc (strlen (current_parsed_root->directory) + sizeof (CVSROOTADM)
+    fname = xmalloc (strlen (current_parsed_root->directory)
+		     + sizeof (CVSROOTADM)
 		     + sizeof (CVSROOTADM_HISTORY) + 3);
     (void) sprintf (fname, "%s/%s/%s", current_parsed_root->directory,
 		    CVSROOTADM, CVSROOTADM_HISTORY);
 
     /* turn off history logging if the history file does not exist */
+    /* FIXME:  This should check for write permissions instead.  This way,
+     * O_CREATE could be added back into the call to open() below and
+     * there would be no race condition involved in log rotation.
+     *
+     * Note that the new method of turning off logging would be either via
+     * the CVSROOT/config file (probably the quicker method, but would need
+     * to be added, or at least checked for, too) or by creating a dummy
+     * history file with 0444 permissions.
+     */
     if (!isfile (fname))
     {
 	logoff = 1;
@@ -733,7 +762,12 @@ history_write (type, update_dir, revs, name, repository)
 		 CLIENT_SERVER_STR, fname);
     if (noexec)
 	goto out;
-    fd = CVS_OPEN (fname, O_WRONLY | O_APPEND | O_CREAT | OPEN_BINARY, 0666);
+
+    if (!history_lock (current_parsed_root->directory))
+	/* history_lock() will already have printed an error on failure.  */
+	goto out;
+
+    fd = CVS_OPEN (fname, O_WRONLY | O_APPEND | OPEN_BINARY, 0666);
     if (fd < 0)
     {
 	if (! really_quiet)
@@ -770,7 +804,7 @@ history_write (type, update_dir, revs, name, repository)
 		if (save_cwd (&cwd))
 		    error_exit ();
 
-		if ( CVS_CHDIR (pwdir) < 0 || (homedir = xgetwd ()) == NULL)
+		if (CVS_CHDIR (pwdir) < 0 || (homedir = xgetwd ()) == NULL)
 		    homedir = pwdir;
 
 		if (restore_cwd (&cwd, NULL))
@@ -877,6 +911,7 @@ history_write (type, update_dir, revs, name, repository)
 	error (1, errno, "cannot close history file: %s", fname);
     free (workdir);
  out:
+    clear_history_lock ();
     free (fname);
 }
 
@@ -890,9 +925,13 @@ save_user (name)
 {
     if (user_count == user_max)
     {
-	user_max += USER_INCREMENT;
-	user_list = (char **) xrealloc ((char *) user_list,
-					(int) user_max * sizeof (char *));
+	user_max = xsum (user_max, USER_INCREMENT);
+	if (size_overflow_p (xtimes (user_max, sizeof (char *))))
+	{
+	    error (0, 0, "save_user: too many users");
+	    return;
+	}
+	user_list = xrealloc (user_list, xtimes (user_max, sizeof (char *)));
     }
     user_list[user_count++] = xstrdup (name);
 }
@@ -920,12 +959,18 @@ save_file (dir, name, module)
 
     if (file_count == file_max)
     {
-	file_max += FILE_INCREMENT;
-	file_list = (struct file_list_str *) xrealloc ((char *) file_list,
-						   file_max * sizeof (*fl));
+	file_max = xsum (file_max, FILE_INCREMENT);
+	if (size_overflow_p (xtimes (file_max, sizeof (*fl))))
+	{
+	    error (0, 0, "save_file: too many files");
+	    return;
+	}
+	file_list = xrealloc (file_list, xtimes (file_max, sizeof (*fl)));
     }
     fl = &file_list[file_count++];
-    fl->l_file = cp = xmalloc (strlen (dir) + strlen (name) + 2);
+    fl->l_file = cp = xmalloc (dir ? strlen (dir) : 0
+			       + name ? strlen (name) : 0
+			       + 2);
     fl->l_module = module;
 
     if (dir && *dir)
@@ -961,9 +1006,13 @@ save_module (module)
 {
     if (mod_count == mod_max)
     {
-	mod_max += MODULE_INCREMENT;
-	mod_list = (char **) xrealloc ((char *) mod_list,
-				       mod_max * sizeof (char *));
+	mod_max = xsum (mod_max, MODULE_INCREMENT);
+	if (size_overflow_p (xtimes (mod_max, sizeof (char *))))
+	{
+	    error (0, 0, "save_module: too many modules");
+	    return;
+	}
+	mod_list = xrealloc (mod_list, xtimes (mod_max, sizeof (char *)));
     }
     mod_list[mod_count++] = xstrdup (module);
 }
@@ -1031,7 +1080,7 @@ fill_hrec (line, hr)
 
 
 #ifndef STAT_BLOCKSIZE
-#if HAVE_ST_BLKSIZE
+#if HAVE_STRUCT_STAT_ST_BLKSIZE
 #define STAT_BLOCKSIZE(s) (s).st_blksize
 #else
 #define STAT_BLOCKSIZE(s) (4 * 1024)
@@ -1362,6 +1411,8 @@ select_hrec (hr)
 		    if (within (cp, cp2))
 		    {
 			hr->mod = fl->l_module;
+			if (cmpfile != NULL)
+			    free (cmpfile);
 			break;
 		    }
 		    if (cmpfile != NULL)
@@ -1524,6 +1575,7 @@ report_hrecs ()
 		break;
 	    case 'W':
 	    case 'U':
+	    case 'P':
 	    case 'C':
 	    case 'G':
 	    case 'M':
