@@ -20,6 +20,7 @@
  */
 
 #include "cvs.h"
+#include <assert.h>
 
 int
 No_Difference (finfo, vers)
@@ -76,6 +77,7 @@ No_Difference (finfo, vers)
 
 	/* update the entdata pointer in the vers_ts structure */
 	p = findnode (finfo->entries, finfo->file);
+	assert (p);
 	vers->entdata = p->data;
 
 	ret = 0;
